/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Incanvas., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import net.hoopajoo.swingx.*;
import java.io.*;

// Experimental cached rooms handler
class CheeseRendererCached extends CheeseRendererBasic {
	// Gridded overlays, TODO: make this a gui option
	HashMap memorytiles = new HashMap();
	boolean renderer_cached_alpha = false;
	protected Vector warplines = new Vector();

	class RenderCacheItem extends java.lang.Object implements java.io.Serializable {
		BufferedImage	image;
		// We need to update this tile if *any* mtimes change, so we store them
		// in this array.  this way we can cache warp lines, even if something
		// happens (like room loading) to the connected rooms, we'll be right
		// the array is the mtimes for each exit with element 0 being this rooms
		// mtime
		long[]		mtimes;
		// A vector that describes the global x/y/z position to draw a line from/to
		// so it's an array of 6 values [x,y,z,x2,y2,z2,off_x,off_y], 
		// plus 2 for offset info (based on tile size
		// because at warp line draw time we don't know if it's a nsewud warp and which exit side
		// to base the line from, the endpoint is always the center though)
		Vector		warplines;

		boolean drew_an_overlay;
	}

	CheeseRendererCached() {
		super();
		// This is useless unless we move to alpha-transparent files instead of
		// 1 bit transparency, which is what most of the current tiles use
		// however if someone makes themes someday then this may be important
		renderer_cached_alpha = CheeseMap.prop_get_boolean( "renderer_cached_alpha" );
	}

	public void prepare() {
		Dimension tile_size = canvas.get_tile_size();
		if( tile_width != tile_size.width ) {
			// Clear all our caches
			memorytiles.clear();
		}

		if( render_grid != canvas.get_render_grid() ) {
			memorytiles.clear();
		}

		// Let super update our properties
		super.prepare();
	}

	// This function draws just the map, not player tiles or HUD data
	// which may change without the map data moving
	public void draw_map( BufferedImage buffer ) {
		screencache = buffer;

		// We mark the sync now so that if there are changed while we
		// are rendering, we will get them
		int rend_x = origin_x,
			rend_y = origin_y,
			rend_z = origin_z;

		int cols = canvas.getWidth() / tile_width;
		int rows = canvas.getHeight() / tile_height;

		CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
		CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );
			
		// Figure our bounding box
		int start_x = rend_x;
		int start_y = rend_y;
		int start_z = rend_z;

		int end_x = start_x + cols;
		int end_y = start_y + rows;
		int end_z = rend_z + 1;

		warplines.clear();
		Graphics2D sg = (Graphics2D)buffer.getGraphics();
			
		// Make a slightly larger font
		sg.setFont( new Font( null, Font.BOLD, 12 ) );
			
		sg.setColor( new Color( 50, 50, 50 ) );
		sg.fillRect( 0, 0, canvas.getWidth(), canvas.getHeight() );
			
		Vector textnodes = new Vector();
		int x, y, z, local_x, local_y;

		for( z = 1; z >= 0; z-- ) {
			for( x = 0; x <= cols; x++ ) {
				local_x = x * tile_width;
				for( y = 0; y <= rows; y++ ) {
					local_y = y * tile_height;
					int global_x = start_x + x;
					int global_y = start_y + y;
					int global_z = start_z + z;

					CheeseRoomData room = CheeseMap.map.get_room( global_x, global_y, global_z );
					if( room != null ) {
						RenderCacheItem thisroom = get_cached_tile( room );
						sg.drawImage( thisroom.image, local_x, local_y, null );

						warplines.addAll( thisroom.warplines );

						if( global_z == rend_z ) {
							// render notes text
							if( room.notes != null && (! room.notes.equals( "" ) ) ) {
								if( room.notes.length() < canvas.get_render_notes_length() && canvas.get_render_notes() ) {
									textnodes.add( room );
								}else{
									if( ! thisroom.drew_an_overlay ) {
										// Render a note tile if notes are here
										CheeseLab.dmesg( 3, "Doing default notes overlay" );
										sg.drawImage( get_image( "overlay_notes" ), local_x, local_y, null );
									}
								}
							}
							
						}else{
							sg.drawImage( get_image( "overlay_blacken" ), local_x, local_y, null );
						}
						
						// Selected
						if( canvas.room_is_selected( room.roomid ) ) {
							CheeseLab.dmesg( 3, "Drawing selected" );
							sg.drawImage( get_image( "selected" ), local_x, local_y, null );   
						}
						
						if( room.roomid == CheeseMap.edit_roomid )
							sg.drawImage( get_image( "infobox" ), local_x, local_y, null );

					}
				}
			}
		}

		// warp lines
		sg.setColor( new Color( 255, 0, 0 ) );
		for( int i = 0; i < warplines.size(); i++ ) {
			int[] points = (int[])warplines.get( i );
			// only on this zplane
			if( points[ 2 ] == rend_z ) {
				Point pstart = canvas.global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
						rend_x, rend_y, rend_z );
				Point pend = canvas.global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
						rend_x, rend_y, rend_z );

			
				sg.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
					pend.x + tile_width / 2, pend.y + tile_height / 2 );
			}
		}
		
		// Notes
		for( int i = 0; i < textnodes.size(); i++ ) {
			CheeseRoomData room =(CheeseRoomData)textnodes.get( i );
			Point p = canvas.global_to_local_reference( room.x, room.y, room.z, rend_x, rend_y, rend_z );
			local_x = p.x;
			local_y = p.y;
			
			int offset_x = 4;
			int offset_y = tile_height - 4;
			
			if( room.x % 2 == 0 ) {
				offset_y = tile_height / 2;
			}
			
			sg.setColor( Color.BLACK );
			sg.drawString( room.notes, 
				local_x + offset_x, local_y + offset_y );
			sg.setColor( Color.LIGHT_GRAY );
			sg.drawString( room.notes, 
				local_x + offset_x + 1, local_y + offset_y + 1 );
		}

		sg.dispose();
	}

	protected RenderCacheItem get_cached_tile( CheeseRoomData room ) {
		RenderCacheItem i = null;
		long[] newmtimes;
		CheeseRoomData[] check = new CheeseRoomData[ 7 ];

		try {
			i = (RenderCacheItem)memorytiles.get( new Integer( room.roomid ) );
		}catch( Exception e ) {
			e.printStackTrace();
		}

		// We need these to verify the rooms haven't changed, and if they have or
		// we haven't cached yet, to make the newmtimes array, so we have to look
		// them up regardless
		check[ 0 ] = room;
		check[ 1 ] = room.exit_n > 0 ? CheeseMap.map.get_room( room.exit_n ) : null;
		check[ 2 ] = room.exit_s > 0 ? CheeseMap.map.get_room( room.exit_s ) : null;
		check[ 3 ] = room.exit_e > 0 ? CheeseMap.map.get_room( room.exit_e ) : null;
		check[ 4 ] = room.exit_w > 0 ? CheeseMap.map.get_room( room.exit_w ) : null;
		check[ 5 ] = room.exit_u > 0 ? CheeseMap.map.get_room( room.exit_u ) : null;
		check[ 6 ] = room.exit_d > 0 ? CheeseMap.map.get_room( room.exit_d ) : null;

		if( i != null ) {
			// verify mtime
			boolean changed = false;

			for( int tmpcount = 0; tmpcount < 7; tmpcount++ ) {
				if( check[ tmpcount ] != null ) {
					if( check[ tmpcount ].mtime > i.mtimes[ tmpcount ] ) {
						changed = true;
					}
				}
			}

			// need to update
			if( changed )
				i = null;
		}

		if( i == null ) {
			// Make cache tile and save it for next time
			i = new RenderCacheItem();
			i.mtimes = new long[ 7 ];
			i.warplines = new Vector();
			i.drew_an_overlay = false;

			for( int tmpcount = 0; tmpcount < 7; tmpcount++ ) {
				if( check[ tmpcount ] != null ) {
					i.mtimes[ tmpcount ] = check[ tmpcount ].mtime;
				}else{
					i.mtimes[ tmpcount ] = 0;
				}
			}

			i.image = (BufferedImage)canvas.createImage( tile_width, tile_height );

			Graphics2D sg = (Graphics2D)i.image.getGraphics();

			// Now paint the tile, we render everything but players and warp lines
			String tilename = canvas.tilename( room );

			sg.drawImage( get_image( "t_" + tilename, render_grid ), 0, 0, null );
			String exittile;

			exittile = get_exit_tile( room, room.exit_e, room.door_e, "e", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
			exittile = get_exit_tile( room, room.exit_w, room.door_w, "w", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
			exittile = get_exit_tile( room, room.exit_n, room.door_n, "n", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
			exittile = get_exit_tile( room, room.exit_s, room.door_s, "s", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
			exittile = get_exit_tile( room, room.exit_u, room.door_u, "u", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
			exittile = get_exit_tile( room, room.exit_d, room.door_d, "d", i.warplines );
			if( exittile != null ) sg.drawImage( get_image( exittile ), 0, 0, null );
		
			for(int ii = 0; ii < room.overlays.length; ii++ ) {
				if( ! room.overlays[ ii ] .equals( "" ) ) {
					CheeseLab.dmesg( 3, "Overlay: " + room.overlays[ ii ] );
					sg.drawImage( get_image( "overlay_" + room.overlays[ ii ] ), 0, 0, null );
					i.drew_an_overlay = true;
				}
			}
							
			sg.dispose();
			memorytiles.put( new Integer( room.roomid ), i );
		}

		return( i );
	}

	private String get_exit_tile( CheeseRoomData room, int exit, String door, String dir, Vector warps ) {
		String tilename = null;
		
		if( exit == 0 ) {
			tilename = "overlay_noexit_" + dir;
		}else if( exit == -2 ) {
			tilename = "overlay_dt_" + dir;
		}else if( exit == -1 ) {
			tilename = "overlay_unexplored_" + dir;
		}else {
			if( ! door.equals( "" ) ) {
				if( door.startsWith( "-" ) ) {
					tilename = "overlay_door_unlocked_" + dir;
				}else if( door.startsWith( "+" ) ) {
					tilename = "overlay_door_locked_" + dir;
				}else{
					tilename = "overlay_door_" + dir;
				}
			}else if( dir.equals( "u" ) || dir.equals( "d" ) ) {
				tilename = "overlay_exit_" + dir;
			}

			// Check for warping
			CheeseRoomData connect = CheeseMap.map.get_room( exit );

			// calc diff from this room to that room, if it's more than 1
			// than we need to render warp lines
			int roomdiff = 0;
			
			roomdiff += Math.abs( connect.x - room.x );
			roomdiff += Math.abs( connect.y - room.y );
			roomdiff += Math.abs( connect.z - room.z );

			if( roomdiff == 1 ) {
				// Make sure it's in the right direction
				if( dir.equals( "e" ) ) {
					if( room.x - connect.x != -1 )
						roomdiff = 0;
				}else if( dir.equals( "w" ) ) {
					if( room.x - connect.x != 1 )
						roomdiff = 0;
				}else if( dir.equals( "n" ) ) {
					if( room.y - connect.y != 1 )
						roomdiff = 0;
				}else if( dir.equals( "s" ) ) {
					if( room.y - connect.y != -1 )
						roomdiff = 0;
				}else if( dir.equals( "u" ) ) {
					if( room.z - connect.z != -1 )
						roomdiff = 0;
				}else if( dir.equals( "d" ) ) {
					if( room.z - connect.z != 1 )
						roomdiff = 0;
				}
			}

			if( roomdiff != 1 ) {
				//CheeseMap.dmesg( "Evaluating warp lines" );
				if( connect.z == room.z && room.z == origin_z ) {
					// Draw warp lines
					int offset_x = 0;
					int offset_y = 0;
					if( dir.equals( "e" ) ) {
						offset_x = tile_width;
						offset_y = tile_height / 2;
					}else if( dir.equals( "w" ) ) {
						offset_x = 0;
						offset_y = tile_height / 2;
					}else if( dir.equals( "n" ) ) {
						offset_x = tile_width / 2;
						offset_y = 0;
					}else if( dir.equals( "s" ) ) {
						offset_x = tile_width / 2;
						offset_y = tile_height;
					}else if( dir.equals( "u" ) ) {
						offset_x = 3;
						offset_y = 3;
					}else if( dir.equals( "d" ) ) {
						offset_x = tile_width - 3;
						offset_y = tile_height - 3;
					}

					CheeseLab.dmesg( 3, "Adding warplines" );
					
					int[] line = new int[ 8 ];
					line[ 0 ] = room.x;
					line[ 1 ] = room.y;
					line[ 2 ] = room.z;

					line[ 3 ] = connect.x;
					line[ 4 ] = connect.y;
					line[ 5 ] = connect.z;
					
					line[ 6 ] = offset_x;
					line[ 7 ] = offset_y;

					warps.add( line );
				}else{
					// Change tile to "overlay_warp"
					// because of diff z planes (connect is not visible)
					tilename = "overlay_warp_" + dir;
				}
			}
		}

		return( tilename );
	}
}
